"""
A resolver based on the Mycroft Adapt intent parser.
"""


import pdb
import logging

import re
from time import time
from ast import Str, Expr, parse

from adapt.intent import IntentBuilder
from adapt.engine import IntentDeterminationEngine

from .base import BaseResolver


class AdaptResolver(BaseResolver):

    def __init__(self, debug):
        self._debug = debug
        self._pifs = {}
        super().__init__(debug)
        self._ad_test = re.compile(r'^add_adapt_handler(\s.+)?')
        self._eng = IntentDeterminationEngine()
        return

    def resolve(self, o_node):
        if 'adapt_r' in self._debug: pdb.set_trace()
        if not self.is_adapt(o_node): return False
        a_node = None
        head = o_node.get('head')
        h_match = self._ad_test.match(head)

        if h_match:
            a_node = self.add_handler(o_node)
            assert a_node is None, f"None expected but got {type(a_node)}"

        else:
            a_node = self.nl_call(o_node)
        return a_node

    def add_handler(self, data):
        engine = self._eng
        ignores = ['intent_parts', 'body']
        data = self._node_to_list(data, ignore_first=ignores)
        if not len(data) == 2:
            raise ValueError(f"Expected 2")
        matches = data[0].get("intent_parts")
        assert all([isinstance(m, Str) for m in matches]), f"Expected all strings"
        itt_parts = [m.s for m in matches]
        body = data[1].get("body")
        body = [Expr(a_n) if not type(a_n) in self._stmt_types else a_n for a_n in body]
        th = f"ad_{abs(hash(time()))}"
        #self._ad_hashes[match] = ad_h
        ib = IntentBuilder(f'{th}')
        ipn_rx = re.compile(r'P<\w+>.')

        for part in itt_parts:
            [
                ib.optionally(ipn[2:-2].partition('_')[-1])
                if 'opt_' in ipn else
                ib.require(ipn[2:-2].partition('_')[-1])
                if 'req_' in ipn else None
                for ipn in ipn_rx.findall(part)
            ]
            part = part.replace('opt_', '', 1).replace('req_', '', 1)
            engine.register_regex_entity(part)
        itt = ib.build()
        engine.register_intent_parser(itt)
        pif = body
        self._pifs[th] = pif
        return None

    def nl_call(self, data):
        head = data.get("head")
        body = data.get("body", [])
        itt = context = p_body = None
        engine = self._eng

        if body:
            # preprocess body
            p_body = self._node_to_list(data)
        intents = [i for i in engine.determine_intent(head)]
        if not intents: return False
        itt = intents[0]
        bit = itt.get('intent_type')
        pif = self._pifs.get(bit)
        if not pif: return False
        assert isinstance(pif, list), f'Unable to resolve `{head}`'
        del itt['intent_type']
        del itt['confidence']
        itta = parse(f'nl_ctx={itt}').body[0]
        node = [itta] + pif
        return node

    def is_adapt(self, node):
        # Just a stub until a way to detect
        return True



#! /usr/bin/env python3
"""

"""


import pdb
import traceback as tb

from ast import *
from os.path import exists
import re
import sys

from .omp import OMP
from .resolvers import Resolver


class AttribProxy:
    """Generate access via attributes"""

    # TODO: move to utils

    def __init__(self, data):
        self._data = data
        # self._func = func
        # self._args = args
        # self._kwds = kwds
        return

    def __getattr__(self, attr):
        data = self._data
        if isinstance(data, dict):
            return data[attr]
        if callable(data):
            return data(attr)
        raise AttributeError("Unable to process {attr}")


class Neulang:
    def __init__(self, debug):
        self._debug = debug
        ns = {}
        self._ns = ns
        self._resolver = ns["RESOLVER"] = Resolver(debug)
        self._ast = None
        self._src = None
        self._latest = None
        self._org_path = None
        return

    def load_astir(self, af=""):
        ns = {}
        if not exists(af):
            return
        am = self.load(af)
        am = compile(am, af, "exec")
        exec(am, ns)
        ap = AttribProxy(ns)
        return ap

    def load(self, path):
        if "core" in self._debug:
            pdb.set_trace()
        if not exists(path):
            raise OSError(f"{path} not found")
        success = None

        with open(path) as po:
            mod = po.read()

            try:
                success = self.loads(mod, path)

            except Exception as e:
                tb.print_exc()
                sys.exit(1)
        return success

    def loads(self, text, sf="<module>"):
        """Convert org text to AST object"""
        ompo = OMP()
        ompo.loads(text, self._org_path)
        self._ast = self._resolver.resolve(ompo.tree)
        self._src = sf
        self._latest = False
        return True

    def eval(self, mode="exec"):
        if self._latest:
            return
        rv = None

        if mode == "exec":
            exec(compile(self._ast, self._src, "exec"), self._ns)

        elif mode == "eval":
            ast = self._ast
            if isinstance(ast, Module):
                ast = ast.body[0]
                ast = Expression(ast.value if isinstance(ast, Expr) else ast)
            rv = eval(compile(ast, self._src, "eval"), self._ns)
        self._latest = True
        return rv

    def reset(self):
        self._ast = self._ns = None
        self._latest = True
        return

    def org_path(self, op):
        if not op or not isinstance(op, list):
            return False
        self._org_path = op
        return True

    @property
    def ns(self):
        return self._ns


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

"""
Neulang: Executable org-formatted pseudocode embedded in Python.
Usage:
    neu [options] [<file>|(-c <command>)|-]  [(-d <areas>)] [(-o <org-path>)] [-m=<modules>]

-v, --version   Print version info.
-i              Open interpreter after executing command. [TODO]
-o, --org-path  Internal traversal path as list of regexes or indices, using sed format (works with <file>).
-m              List of modules to import, using bash format. [TODO]
-d, --debug     Enable debugging mode for areas (astir|resolver|pyexpr_r|regex_r|core|main|base|air_*|...).
"""


import traceback as tb
import pdb

import sys
import json
from os.path import exists, join, dirname

try:
    from docopt import docopt

except:
    # likely just installing
    pass

from .core import Neulang


__all__ = ("__meta__", "__version__", "run")
meta_file = join(dirname(__file__), "meta.json")
if not exists(meta_file):
    open(meta_file, "w").write("{}")
__meta__ = json.load(open(meta_file))
__author__ = __meta__.get("author", "skeledrew")
__version__ = __meta__.get("version", "0.1.0")


def _run():
    # catch assert errors

    try:
        run()

    except AssertionError as ae:
        tb.print_exc()
        pdb.post_mortem()
        sys.exit(1)
    return


def run():
    opt = docopt(__doc__)
    debug = opt.get("<areas>") if opt.get("--debug") else ""
    if opt.get("--debug") and not debug:
        debug = "main"
    if "main" in debug:
        pdb.set_trace()
    src_file = opt.get("<file>") or ""
    q_cmd = opt.get("<command>") if opt.get("-c") else None
    org_path = opt.get("<org-path>") if opt.get("--org-path") else ""
    org_path = (
        org_path.split(org_path[1])[1:]
        if org_path.startswith("s") and len(org_path) > 2
        else None
    )
    neulang = Neulang(debug)

    if not any(list(opt.values())) or opt.get('-i'):
        # nothing passed
        level = 1
        accum = []
        i_char = '\\'
        d_char = '/'
        q_words = ['*quit*', '*exit*']
        print(f"Neulang {__version__}")
        print(f'Use "{i_char}" or "{d_char}" to indent or deindent lines.')
        print(f'Evaluate code by entering an empty line after.')
        print(f'Type "{q_words[0]}" or "{q_words[1]}" to quit.\n')

        while True:
            # basic interpreter
            rv = ''
            stars = '*' * level
            prompt = f"{stars} "

            try:
                inp = input(prompt).strip()

            except KeyboardInterrupt:
                print('\nKeyboardInterrupt')
                continue

            except EOFError:
                print('')
                break

            if inp in q_words:
                break

            if accum and not inp:

                try:
                    neulang.loads('\n'.join(accum))

                    try:
                        rv = neulang.eval("eval")
                        if rv is None: rv = ''

                    except (SyntaxError, TypeError):
                        # NB: catching TypeError might mask pertinent errors
                        neulang.eval()

                    else:
                        print(rv)

                except KeyboardInterrupt:
                    continue

                except Exception as e:
                    tb.print_exc()

                finally:
                    level = 1
                    accum = []
                continue

            if inp is i_char or inp.endswith(i_char):
                inp = inp.rstrip(i_char)
                level += 1

            elif inp is d_char or inp.endswith(d_char):
                inp = inp.rstrip(d_char)
                level -= 1
                if level < 1: level = 1
            if inp: accum.append(f'{prompt}{inp}')

    elif src_file:
        # source file
        neulang.org_path(org_path)
        neulang.load(src_file)
        neulang.eval()

    elif q_cmd:
        # quoted string command
        if not q_cmd.startswith("* "):
            q_cmd = f"* {q_cmd}"
        neulang.loads(q_cmd)
        neulang.eval()

    elif opt.get("--version", None):
        print(f"Neulang {__version__}")
    return


if __name__ == "__main__":
    run()
